# -*- coding: utf-8 -*-
"""
Ochrona-cli
:author: ascott
"""


class OchronaException(Exception):
    pass


class OchronaAPIException(OchronaException):
    pass


class OchronaFileException(OchronaException):
    pass


class OchronaImportException(OchronaException):
    pass



# -*- coding: utf-8 -*-
"""
Ochrona-cli
:author: ascott
"""

import json
import os
import platform
import re
import sys
import yaml
from typing import List, Optional


class OchronaConfig:
    """
    Config for running ochrona cli.
    """

    _api_key: Optional[str] = None
    _debug: bool = False
    _silent: bool = False
    _dir: Optional[str] = None
    _file: Optional[str] = None
    _python_version: Optional[str] = None
    _report_type: Optional[str] = None
    _report_location: Optional[str] = None
    _api_url: Optional[str] = None
    _alert_api_url: Optional[str] = None
    _exit: bool = False
    _ignore: Optional[str] = None
    _include_dev: bool = False
    _color_output: bool = True

    _project_name: Optional[str] = None
    _alert_config: Optional[str] = None

    REPORT_OPTIONS: List[str] = ["BASIC", "FULL", "JSON", "XML"]
    DEFAULT_SL_API_URL: str = "https://api.ochrona.dev/python/analyze"
    DEFAULT_ALERT_API_URL: str = "https://api.ochrona.dev/alerts/project-alerts"

    ALERT_ADDRESS_PATTERN: str = r"^[A-Za-z0-9\.\+_-]+@[A-Za-z0-9\._-]+\.[a-zA-Z]*$"

    def __init__(self, **kwargs):
        self.get_config(**kwargs)
        valid = self._validate()
        if not valid[0]:
            sys.exit(valid[1])
        self._detect_runtime_attributes()

    def get_config(self, **kwargs):
        """
        Returns a config object for running dependency vulnerability scan.

        Config precedence is:
            1. Command line
            2. Environment Variables
            2. Config
        :param kwargs:
        :return:
        """

        # check command line args
        self._api_key = kwargs.get("api_key")
        self._debug = kwargs.get("debug", False)
        self._silent = kwargs.get("silent", False)
        self._dir = kwargs.get("dir")
        self._file = kwargs.get("file")
        self._report_type = kwargs.get("report_type")
        location = kwargs.get("report_location")
        if location == ".":
            self._report_location = os.getcwd()
        else:
            self._report_location = location
        self._exit = kwargs.get("exit", False)
        self._ignore = kwargs.get("ignore")
        self._include_dev = kwargs.get("include_dev", False)
        self._project_name = kwargs.get("project_name")
        if "alert_config" in kwargs and kwargs.get("alert_config") is not None:
            self._alert_config = json.loads(kwargs.get("alert_config"))

        # check runtime environment
        sys_type = platform.system()
        if sys_type == "Windows":
            self._color_output = False

        # check environment variables
        self._api_key = (
            self._api_key if self._api_key else os.environ.get("OCHRONA_API_KEY", None)
        )
        self._debug = (
            self._debug
            if self._debug
            else os.environ.get("OCHRONA_DEBUG_LOGGING", False)
        )
        self._api_url = os.environ.get(
            "OCHRONA_API_URL", OchronaConfig.DEFAULT_SL_API_URL
        )
        self._alert_api_url = os.environ.get(
            "OCHRONA_ALERT_API_URL", OchronaConfig.DEFAULT_ALERT_API_URL
        )
        self._ignore = (
            self._ignore
            if self._ignore
            else os.environ.get("OCHRONA_IGNORED_VULNS", None)
        )
        if self._ignore:
            self._ignore = self._ignore.split(",")

        # check config .ochrona.yml
        try:
            with open(".ochrona.yml", "r") as stream:
                yaml_loaded = yaml.safe_load(stream)
                if yaml_loaded:
                    self._api_key = yaml_loaded.get("api_key", self._api_key)
                    self._api_url = yaml_loaded.get("api_url", self._api_url)
                    self._alert_api_url = yaml_loaded.get(
                        "alert_url", self._alert_api_url
                    )
                    self._debug = yaml_loaded.get("debug", self._debug)
                    self._silent = yaml_loaded.get("silent", self._silent)
                    self._dir = yaml_loaded.get("dir", self._dir)
                    self._file = (
                        yaml_loaded["file"] if "file" in yaml_loaded else self._file
                    )
                    self._report_type = (
                        yaml_loaded["report_type"]
                        if "report_type" in yaml_loaded
                        else self._report_type
                    )
                    self._report_location = (
                        yaml_loaded["report_location"]
                        if "report_location" in yaml_loaded
                        else self._report_location
                    )
                    self._exit = yaml_loaded.get("exit", self._exit)
                    self._ignore = yaml_loaded.get("ignore", self._ignore)
                    self._include_dev = yaml_loaded.get("include_dev", self._ignore)
                    self._color_output = yaml_loaded.get(
                        "color_output", self._color_output
                    )
                    # Project and DADA configuration
                    self._project_name = yaml_loaded.get(
                        "project_name", self._project_name
                    )
                    if yaml_loaded.get("alert_config") is not None:
                        self._alert_config = {}
                        self._alert_config["alerting_addresses"] = yaml_loaded.get(
                            "alert_config"
                        ).get("alerting_addresses")
                        self._alert_config["alerting_rules"] = yaml_loaded.get(
                            "alert_config"
                        ).get("alerting_rules")
        except IOError:
            pass

    def _validate(self):
        """
        Validates all required values are present.
        """
        if not self._api_key:
            return (False, "Missing config value `api_key`")
        if self._report_type not in self.REPORT_OPTIONS:
            return (False, f"Unknown report type specified in {self._report_type}")
        if self._alert_config is not None:
            if "alerting_addresses" not in self._alert_config:
                return (False, "Missing alerting_addresses in DADA alert config")
            emails = self._alert_config.get("alerting_addresses").split(",")
            if len(emails) <= 0:
                return (
                    False,
                    "alerting_addresses in DADA alert config was found empty",
                )
            for email in emails:
                if not re.match(self.ALERT_ADDRESS_PATTERN, email):
                    return (False, f"Invalid email address {email} found in config")
        return (True, None)

    def _detect_runtime_attributes(self):
        """
        Detects details about the runtime and updates the config.
        """
        self._python_version = ".".join([str(i) for i in sys.version_info][0:3])

    @property
    def api_key(self):
        return self._api_key

    @property
    def debug(self):
        return self._debug

    @property
    def silent(self):
        return self._silent

    @property
    def dir(self):
        return self._dir

    @property
    def file(self):
        return self._file

    @property
    def report_type(self):
        return self._report_type

    @property
    def report_location(self):
        return self._report_location

    @property
    def api_url(self):
        return self._api_url

    @property
    def alert_api_url(self):
        return self._alert_api_url

    @property
    def exit(self):
        return self._exit

    @property
    def ignore(self):
        return self._ignore

    @property
    def include_dev(self):
        return self._include_dev

    @property
    def color_output(self):
        return self._color_output

    @property
    def project_name(self):
        return self._project_name

    @property
    def alert_config(self):
        return self._alert_config



# -*- coding: utf-8 -*-
"""
Ochrona-cli
:author: ascott
"""

import os

from typing import Any, Dict, IO, Optional, List, Type, TextIO

from ochrona.config import OchronaConfig
from ochrona.const import (
    CONDA_ENVIRONMENT,
    CONSTRAINTS_TXT,
    PIPFILE_LOCK,
    POETRY_LOCK,
    REQUIREMENTS_TXT,
    SETUP_PY,
    SUPPORTED_DEPENDENCY_FILE_PATTERNS,
    TOX_INI,
)
from ochrona.exceptions import OchronaFileException
from ochrona.logger import OchronaLogger
from ochrona.parsers import Parsers

Path: Optional[Type] = None
try:
    from pathlib import Path
except ImportError:
    pass


def rfind_all_dependencies_files(
    logger: OchronaLogger,
    directory: Optional[str] = None,
    file_obj: Optional[IO] = None,
) -> List[str]:
    """
    Recursively searches for dependency files to analyze

    :param logger: A configured `OchronaFormatter` instance
    :param directory: str - starting directory (optional)
    :param file_obj: A specified file to use
    :return: list - list of paths for files to analyze
    """
    if not directory:
        directory = os.getcwd()

    files = []

    if file_obj:
        files.append(file_obj.name)
    else:
        if Path:
            for filename in Path(directory).glob(
                SUPPORTED_DEPENDENCY_FILE_PATTERNS[REQUIREMENTS_TXT]
            ):
                logger.debug(f"Found matching requirements*.txt file at {filename}")
                files.append(filename)
            for filename in Path(directory).glob(
                SUPPORTED_DEPENDENCY_FILE_PATTERNS[PIPFILE_LOCK]
            ):
                logger.debug(f"Found matching pipfile.lock file at {filename}")
                files.append(filename)
            for filename in Path(directory).glob(
                SUPPORTED_DEPENDENCY_FILE_PATTERNS[POETRY_LOCK]
            ):
                logger.debug(f"Found matching poetry.lock file at {filename}")
                files.append(filename)
            for filename in Path(directory).glob(
                SUPPORTED_DEPENDENCY_FILE_PATTERNS[SETUP_PY]
            ):
                logger.debug(f"Found matching setup.py file at {filename}")
                files.append(filename)
            for filename in Path(directory).glob(
                SUPPORTED_DEPENDENCY_FILE_PATTERNS[CONDA_ENVIRONMENT]
            ):
                logger.debug(f"Found matching conda environment.yml file at {filename}")
                files.append(filename)
            for filename in Path(directory).glob(
                SUPPORTED_DEPENDENCY_FILE_PATTERNS[TOX_INI]
            ):
                logger.debug(f"Found matching conda tox.ini file at {filename}")
                files.append(filename)
            for filename in Path(directory).glob(
                SUPPORTED_DEPENDENCY_FILE_PATTERNS[CONSTRAINTS_TXT]
            ):
                logger.debug(f"Found matching conda constraints.txt file at {filename}")
                files.append(filename)

    if not files:
        raise OchronaFileException("No dependencies files found")

    return files


def parse_to_payload(
    logger: OchronaLogger, file_path: str, config: OchronaConfig
) -> Dict[str, Any]:
    """
    Parses a requirements.txt type file or Pipefile.lock into a JSON payload.
    :param logger: A configured `OchronaLogger` instance
    :param file_path: path to file
    :param config: An instance of `OchronaConfig`
    :return: JSON payload
    """
    dependencies = []
    parsers = Parsers()
    if os.path.basename(file_path).lower() == PIPFILE_LOCK.lower():
        dependencies = parsers.pipfile.parse(
            file_path=file_path, include_dev=config.include_dev
        )
    elif os.path.basename(file_path).lower() == POETRY_LOCK.lower():
        dependencies = parsers.poetry.parse(
            file_path=file_path, include_dev=config.include_dev
        )
    elif os.path.basename(file_path).lower() == SETUP_PY.lower():
        dependencies = parsers.setup.parse(
            file_path=file_path, include_dev=config.include_dev
        )
    elif os.path.basename(file_path).lower() == CONDA_ENVIRONMENT.lower():
        dependencies = parsers.conda.parse(file_path=file_path)
    elif os.path.basename(file_path).lower() == TOX_INI.lower():
        dependencies = parsers.tox.parse(file_path=file_path)
    elif os.path.basename(file_path).lower() == CONSTRAINTS_TXT.lower():
        dependencies = parsers.constraints.parse(file_path=file_path)
    else:
        dependencies = parsers.requirements.parse(file_path=file_path)
    logger.debug(f"Discovered dependencies: {dependencies}")
    if config.project_name is not None:
        return {"dependencies": dependencies, "project_name": config.project_name}
    else:
        return {"dependencies": dependencies}



# -*- coding: utf-8 -*-
"""
Ochrona-cli
:author: ascott
"""

import requests

from ochrona.logger import OchronaLogger


class PYPIAPIClient:
    def __init__(self, logger: OchronaLogger):
        self.logger = logger

    def latest_version(self, package: str) -> str:
        """
        Calls pypi to fetch the latest version of a package.

        :param package: package name
        :return: str - latest version
        """
        url = "https://pypi.org/pypi/{}/json".format(package)
        response = requests.request("GET", url)
        if response.status_code == 200:
            json = response.json()
            if json:
                return json.get("info", {}).get("version", "")
        return ""



# -*- coding: utf-8 -*-
"""
Ochrona-cli
:author: ascott
"""

PIPFILE_LOCK = "Pipfile.lock"
POETRY_LOCK = "poetry.lock"
SETUP_PY = "setup.py"
REQUIREMENTS_TXT = "requirements.txt"
CONSTRAINTS_TXT = "constraints.txt"
CONDA_ENVIRONMENT = "environment.yml"
TOX_INI = "tox.ini"

SUPPORTED_DEPENDENCY_FILE_PATTERNS = {
    REQUIREMENTS_TXT: "**/*requirements*.txt",
    CONSTRAINTS_TXT: "**/*constraints*.txt",
    PIPFILE_LOCK: "**/*Pipfile.lock",
    POETRY_LOCK: "**/*poetry.lock",
    SETUP_PY: "**/*setup.py",
    CONDA_ENVIRONMENT: "**/environment.yml",
    TOX_INI: "**/tox.ini",
}

INVALID_REQUIREMENTS_LINES = [
    "#",
    "-i",
    "-f",
    "-Z",
    "--index-url",
    "--extra-index-url",
    "--find-links",
    "--no-index",
    "--allow-external",
    "--allow-unverified",
    "--always-unzip",
]

TOX_LINKED_REQUIREMENTS = "-r"
INVALID_TOX_LINES = ["{"]

INVALID_SPECIFIERS = {"<", ">", "!", "~"}
EQUALS_SPECIFIER = "=="



from typing import List

import toml
from ochrona.exceptions import OchronaFileException


class PoetryFile:
    @staticmethod
    def parse(file_path: str, include_dev: bool = False) -> List[str]:
        """
        Parses a poetry.lock file into a list of requirements.

        :param file_path: poetry.lock file path
        :return: list<str> list of dependencies ['dependency==semvar']
        """
        dependencies = []
        try:
            with open(file_path) as poetry_lock:
                parsed = toml.load(poetry_lock)
                for pkg in parsed.get("package", []):
                    if pkg.get("category") == "main":
                        dependencies.append(f"{pkg.get('name')}=={pkg.get('version')}")
                    elif pkg.get("category") == "dev" and include_dev:
                        dependencies.append(f"{pkg.get('name')}=={pkg.get('version')}")
            return dependencies
        except OSError as ex:
            raise OchronaFileException(f"OS error when parsing {file_path}") from ex
        except toml.decoder.TomlDecodeError as ex:  # type: ignore[attr-defined]
            raise OchronaFileException(
                f"Could not parse {file_path} - is TOML valid?"
            ) from ex



# -*- coding: utf-8 -*-
"""
Ochrona-cli
:author: ascott
"""

import json
import sys
import requests
from typing import Any, Dict, Optional, Union

from ochrona import __version__
from ochrona.config import OchronaConfig
from ochrona.exceptions import OchronaAPIException
from ochrona.logger import OchronaLogger


class OchronaAPIClient:
    def __init__(self, logger: OchronaLogger, config: OchronaConfig):
        self._api_key = config.api_key
        self.logger = logger
        self._url = config.api_url
        self._alert_api_url = config.alert_api_url

    def analyze(self, payload: Optional[str] = None) -> Dict[str, Any]:
        """
        Calls the analysis API endpoint and returns results.

        :param payload: json payload
        :return: dict
        """
        response = requests.request(
            "POST", self._url, headers=self._generate_headers(), data=payload
        )
        if response.status_code > 300:
            self.logger.debug(f"Unexpected Response from {self._url} - {response.text}")
            self._error_response_handler(response.status_code)
        return self._response_handler(response)

    def update_alert(self, config: Union[Any, OchronaConfig]) -> Dict[str, Any]:
        """
        Calls the Alert Registration API to update project alert settings.

        :param payload: json payload
        :return: dict
        """
        response = requests.request(
            "POST",
            self._alert_api_url,
            headers=self._generate_headers(),
            data=self._alert_payload_handler(config),
        )

        if response.status_code > 300:
            self.logger.debug(
                f"Unexpected Response from {self._alert_api_url} - {response.text}"
            )
            self._error_response_handler(response.status_code)
        return self._response_handler(response)

    def _generate_headers(self) -> Dict[str, str]:
        return {
            "Content-Type": "application/json",
            "x-api-key": f"{self._api_key}",
            "User-Agent": f"OchronaClient/{__version__}/{'.'.join([str(i) for i in sys.version_info][0:3])}",
            "Accept": "*/*",
            "Cache-Control": "no-cache",
            "Host": "api.ochrona.dev",
        }

    def _error_response_handler(self, status_code: int):
        """
        Logs a user friendly message based on the error returned from the server.

        :param status_code: int
        :return:
        """
        if 500 >= status_code >= 400:
            raise OchronaAPIException(
                "Unexpected request sent for analysis. "
                "Please report this at https://github.com/ochronasec/ochrona-cli/issues"
            )
        elif status_code >= 500:
            raise OchronaAPIException(
                "Unexpected result from analysis, please try again later. If this persists, "
                "please report this at https://github.com/ochronasec/ochrona-cli/issues"
            )
        else:
            raise OchronaAPIException(
                "Unexpected response from API: {}".format(status_code)
            )

    def _response_handler(self, response: requests.models.Response) -> Dict[str, Any]:
        """
        Returns the parsed json response as a dict.

        :param response: Response object from request
        :return: dict
        """
        try:
            return json.loads(response.text)
        except json.JSONDecodeError as ex:
            raise OchronaAPIException("Error parsing API response") from ex

    def _alert_payload_handler(self, config: OchronaConfig) -> str:
        """
        Parses the config object and sends to the Ochrona Alert Registration API.

        :param config: OchronaConfig instance
        :return: dict
        """
        return json.dumps(
            {
                "project_name": config.project_name,
                "alerting_addresses": config.alert_config.get("alerting_addresses"),
                "alerting_rules": config.alert_config.get("alerting_rules"),
            }
        )



import ast

from typing import List

from ochrona.exceptions import OchronaFileException


class SetupFile:
    @staticmethod
    def parse(file_path: str, include_dev: bool = False) -> List[str]:
        """
        Parses a setup.py into a list of requirements.

        :param file_path: setup.py path
        :return: list<str> list of dependencies ['dependency==semvar']
        """
        setup_func_names = {"setup", "setuptools.setup"}
        try:
            dependencies = []
            with open(file_path) as setuppy:
                tree = ast.parse(setuppy.read())
                for body in tree.body:
                    if isinstance(body, ast.Expr):
                        if body.value.func.id in setup_func_names:  # type: ignore[attr-defined]
                            for kw in body.value.keywords:  # type: ignore[attr-defined]
                                if kw.arg == "install_requires":
                                    dependencies += [arg.s for arg in kw.value.elts]
                                elif kw.arg == "tests_require" and include_dev:
                                    dependencies += [arg.s for arg in kw.value.elts]
            return dependencies
        except OSError as ex:
            raise OchronaFileException(f"OS error when parsing {file_path}") from ex
        except AttributeError as ex:
            raise OchronaFileException(f"AST error when parsing {file_path}") from ex



# -*- coding: utf-8 -*-
"""
Ochrona-cli
:author: ascott
"""

import json
import os
import subprocess
import sys
from typing import Any, Dict, List, Optional, Tuple, Union

from ochrona.const import INVALID_SPECIFIERS, EQUALS_SPECIFIER
from ochrona.exceptions import OchronaImportException
from ochrona.parsers import Parsers
from ochrona.logger import OchronaLogger
from ochrona.pypi_rest_client import PYPIAPIClient
from ochrona.ochrona_rest_client import OchronaAPIClient


class SafeImport:
    def __init__(self, logger: OchronaLogger, client: OchronaAPIClient):
        self.logger = logger
        self.ochrona = client
        self.pypi = PYPIAPIClient(logger=logger)

    def install(self, package: str):
        """
        Calls the analysis API endpoint with a package and if it's safe, imports via pip.

        :param package: a package string with an optional version specified
        :return: bool
        """
        if os.path.isfile(package):
            parsers = Parsers()
            packages = parsers.requirements.parse(package)
            if self._check_package(packages):
                self.logger.info(
                    f"No vulnerabilities found for {', '.join(packages)}, install proceeding."
                )
                self._install_file(package)
            else:
                self.logger.error(
                    f"Import of {', '.join(packages)} aborted due to detected vulnerabilities."
                )
        else:
            if self._check_package(package):
                self.logger.info(
                    f"No vulnerabilities found for {package}, install proceeding."
                )
                self._install(package=package)
            else:
                self.logger.error(
                    f"Import of {package} aborted due to detected vulnerabilities."
                )

    def _check_package(self, package: Union[str, list]) -> bool:
        """
        Calls the analysis API endpoint with a package to see if a package is safe.

        :param package: a package string with an optional version specified
        :return: bool
        """
        vuln_results: Tuple[Union[Any, List[Any]], Union[Any, List[Any]]] = ([], [])
        if isinstance(package, str):
            if any(spec in package for spec in INVALID_SPECIFIERS):
                raise OchronaImportException(
                    f"An invalid specifier was found in {package}, either specify the package without a version or pin to a single version using `name==version`."
                )
            parts = package.split(EQUALS_SPECIFIER)
            if len(parts) == 1:
                package = self._get_most_recent_version(package=package)
            vuln_results = self._parse_ochrona_results(
                self.ochrona.analyze(json.dumps({"dependencies": [package]}))
            )
        elif isinstance(package, list):
            vuln_results = self._parse_ochrona_results(
                self.ochrona.analyze(json.dumps({"dependencies": package}))
            )
        if len(vuln_results[1]) > 0:
            self.logger.info(
                f"A full list of packages that would be installed, included dependencies: {', '.join(vuln_results[0])}"
            )
            self.logger.error(
                f"""Vulerabilities found related to import:\n{''.join([self._format_vulnerability(v) for v in vuln_results[1]])}"""
            )
            return False
        self.logger.info(
            f"A full list of packages to be installed, included dependencies: {', '.join(vuln_results[0])}"
        )
        return True

    def _get_most_recent_version(self, package: str) -> str:
        """
        If a package does not have a version specified we will assume that the latest version
        should be used.

        :param package: a package string without a version
        :return: str
        """
        pypi_response = self.pypi.latest_version(package=package)
        if pypi_response != "":
            return f"{package}{EQUALS_SPECIFIER}{self.pypi.latest_version(package=package)}"
        else:
            self.logger.warn("Unable to reach Pypi to confirm latest version")
            return package

    def _parse_ochrona_results(
        self, results: Dict[str, Any]
    ) -> Tuple[Union[Any, List[Any]], Union[Any, List[Any]]]:
        """
        Parses Ochrona API results and outputs information regarding any confirmed vulnerabilities.
        :param results: API results
        :return: tuple<list, list> - a list of discovered vulnerabilities, a list of dependencies
        """
        return (
            results.get("flat_list") or [],
            results.get("confirmed_vulnerabilities") or [],
        )

    def _install(self, package: str) -> bool:
        """
        Call pip to install a package

        :param package: a package string with a version
        :return: bool
        """
        try:
            subprocess.check_call([sys.executable, "-m", "pip", "install", package])
            return True
        except subprocess.CalledProcessError as ex:
            raise OchronaImportException("Error during pip install") from ex

    def _install_file(self, file_path: str) -> bool:
        """
        Call pip to install a requirements.txt style

        :param file_path: path to requirements file
        :return: bool
        """
        try:
            subprocess.check_call(
                [sys.executable, "-m", "pip", "install", "-r", file_path]
            )
            return True
        except subprocess.CalledProcessError as ex:
            raise OchronaImportException("Error during pip install") from ex

    def _format_vulnerability(self, vulnerability: Dict[str, Any]) -> str:
        """
        Formats a vulnerability to a friendly output

        :param vulnerability: vulnerability results
        :return: str
        """
        return f"""\nVulnerability Detected on package to be installed!
    Package: {vulnerability.get('name')}
    ID: {vulnerability.get('cve_id')}
    Description: {vulnerability.get('description')}
    Ochrona Severity: {vulnerability.get('ochrona_severity_score')}
                """
